#ifndef GEOM_FILE
#define GEOM_FILE

/**
 * The quadratic domain [0,1] x [0,1] is divided into @p procx * @p procy subdomains
   numbered rowise.
   According to my process with rank number @p myid, the vector @p neigh contains the ranks
   of the neighbours t the South, East, North, West.
 * @param[in] myid	my rank
 * @param[in] procx	number of processes in x-direction
 * @param[in] procy	number of processes in y-direction
 * @param[out] neigh	vector of 4 elements containing the ranks of neighbouring processes, -1 indicates a boundary in this direction
 * @param[out] color	red/black coloring of subdomains (0/1)
*/
void IniGeom(const int myid, const int procx, const int procy, int neigh[], int& color);


/**
 * The quadratic domain [0,1] x [0,1] is divided into @p procx * @p procy subdomains
   numbered rowise.
   According to my process with rank number @p myid, the coordinates of the lower
   left corner (@p xl, @p yb) and of the upper right corner (@p xr, @p yt) are generated.
 * @param[in] myid	my rank
 * @param[in] procx	number of processes in x-direction
 * @param[in] procy	number of processes in y-direction
 * @param[out] xl	x-coordinate of left boundary
 * @param[out] xr	x-coordinate of right boundary
 * @param[out] yb	y-coordinate of lower boundary
 * @param[out] yt	y-coordinate of upper boundary
*/
void IniCoord(const int myid, const int procx, const int procy,
	      float& xl, float& xr, float& yb, float& yt);

/**
 * Determines the coordinates of the dicretization nodes of the domain [@p xl, @p xr] x [@p yb, @p yt]
 * which is discretized into @p nx x @p ny intervals.
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[in] xl	x-coordinate of left boundary
 * @param[in] xr	x-coordinate of right boundary
 * @param[in] yb	y-coordinate of lower boundary
 * @param[in] yt	y-coordinate of upper boundary
 * @param[out] xc	coordinate vector of length 2n with x(2*k,2*k+1) as coodinates of node k
*/

void GetCoordsInRectangle(const int nx, const int ny,
	const float& xl, const float& xr, const float& yb, const float& yt,
	float xc[]);
/**
 * Determines the element connectivity of linear triangular elements of a FEM discretization
 * of a rectangle using @p nx x @p ny equidistant intervals for discretization.
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[out] ia	element connectivity matrix with ia(3*s,3*s+1,3*s+2) as node numbers od element s
*/
void GetConnectivityInRectangle(const int nx, const int ny, int ia[]);


/**
 * Copies the values of @p w corresponding to boundary @p ib
 * onto vector s.  South (ib==1), East (ib==2), North (ib==3), West (ib==4).
 * The vector @p s has to be long enough!!
 * @param[in] ib	my local boundary
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[in] w		vector for all nodes of local discretization
 * @param[out] s	short vector with values on boundary @p ib
*/
// GH_NOTE: Absicherung bei s !!
void GetBound(const int ib, const int nx, const int ny, const float w[], float s[]);


/**
 * Computes @p w := @p w + @p s  at the interface/boundary nodes on the
 * boundary @p ib .  South (ib==1), East (ib==2), North (ib==3), West (ib==4)
 * @param[in] ib	my local boundary
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[in,out] w	vector for all nodes of local discretization
 * @param[in] s		short vector with values on boundary @p ib
*/
void AddBound(const int ib, const int nx, const int ny, float w[], const float s[]);


/**
 * Stores the values of vector @p u of subdomain @p myid into a file.
 * The file stores rowise the x- and y- coordinates together with the value from  @p u .
 * The domain [@p xl, @p xr] x [@p yb, @p yt] is discretized into @p nx x @p ny intervals.
 * @param[in] myid	my rank
 * @param[in] name	basename of file name (file name will be extended by the rank number)
 * @param[in] u		local vector
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[in] xl	x-coordinate of left boundary
 * @param[in] xr	x-coordinate of right boundary
 * @param[in] yb	y-coordinate of lower boundary
 * @param[in] yt	y-coordinate of upper boundary
*/
void SaveVectorP(const int myid, const string& name, const float u[],
                 const int nx, const int ny,
                 const float& xl, const float& xr, const float& yb, const float& yt);


/**
      * The 2D domain [xl,xr] x [yb,yt] is discretized into a triangular mesh
        based on @p nx times @p ny equidistant grid of the rectangle.

        The memory for coordinates array @p xc  and for the element connectivity @p ia is
        allocated inside the function.

      * @param[in] nx	number of equidistant intervals in x-direction
      * @param[in] ny   number of equidistant intervals in y-direction
      * @param[in] xl   x-coordinate of left boundary
      * @param[in] xr	x-coordinate of right boundary
      * @param[in] yb	y-coordinate of lower boundary
      * @param[in] yt	y-coordinate of upper boundary
      * @param[out] nnode  number of coordinates
      * @param[out] xc     array[nnode][2] of x- and y-coordinates
      * @param[out] nelem  number of triangular elements
      * @param[out] ia     array[nelem][3] element connectivity for linear triangular elements
*/
void GetMesh(const int nx, const int ny, const float xl, const float xr, const float yb, const float yt,
  int &nnode, float* &xc, int &nelem, int* &ia);


/**
      * Prints the information for a finite element mesh
      *
      * @param[in]  nnode      number coordinates
      * @param[in]  xc[nnode]  array of coordinates
      * @param[in]  nelem      number of elements
      * @param[in]  ndof_e     degrees of freedom per element (= number of geometric points)
      * @param[in]  ia[nelem][ndof_e]  element connectivity array for linear triangular elements
*/

void DebugMesh(const int nnode, const float xc[], const int nelem, const int ndof_e, const int ia[]);

#endif
 
