//		MPI code in C++.
//		See [Gropp/Lusk/Skjellum, "Using MPI", p.33/41 etc.]
//		and  /opt/mpich/include/mpi2c++/comm.h  for details

// main programm for versions  1-3 where
// all data transfer/allocation are still done inside function JacobiSolve_GPU

#include <iostream>
#include <cmath>
#include <ctime>
using namespace std;
#include "mpi.h"

float dscapr_GPU(const int n, const float x[], const float y[]);


int main(int argc, char** argv)
{
//****************  start MPI  ****************************************
    MPI::Init(argc,argv);
    MPI::Intracomm icomm(MPI::COMM_WORLD);
    const int numprocs = icomm.Get_size();
    const int myrank   = icomm.Get_rank();

    if (myrank == 0)
    { cout << "\n There are " << numprocs << " MPI processes running.\n \n"; }
//*********************************************************************

//********   Allocate and init data on  CPU and GPU  ******************
 float *x_h, *y_h; // host data
 float *x_d, *y_d, *s_d; // device data
 float sum;
 const int N = 14000000,
           nBytes = N*sizeof(float);
 const int LOOP = 100;

 int deviceCount;
 cudaGetDeviceCount(&deviceCount);                    // How many GPUs?
 int device_id = myrank % deviceCount;
 cudaSetDevice(device_id);                            // Map MPI-process to a GPU
 cout << "Map MPI (" << myrank << "/" << numprocs << ") onto  device [" <<
          device_id << "/" << deviceCount << "]" << endl;
 icomm.Barrier();                                     // MPI call

 cudaDeviceProp prop;
 cudaGetDeviceProperties (&prop, 0);

 const int blocksize= 4*64,
//           gridsize = ceil( N / static_cast<float>(blocksize)),
           gridsize = prop.multiProcessorCount,
           resBytes = gridsize*sizeof(float);
 dim3 dimBlock(blocksize);
 dim3 dimGrid(gridsize);

 cout << "(" << myrank << "/" << numprocs << "): " <<
       prop.name << " GPU  with " <<
       prop.multiProcessorCount << " Multiprocessors : "     <<
       gridsize << " x " << blocksize << " Threads\n";

 x_h = new float [N];
 y_h = new float [N];
// allocate memory on device
 cudaMalloc((void **) &x_d, nBytes);
 cudaMalloc((void **) &y_d, nBytes);
 cudaMalloc((void **) &s_d, resBytes);

 float ss=0.0;
 for (int i=0; i<N; i++)
   { x_h[i] = (i % 137)+1; y_h[i] = 1.0/x_h[i]; ss+=x_h[i]*y_h[i];}

// copy data:  host --> device
 cudaMemcpy(x_d, x_h, nBytes, cudaMemcpyHostToDevice);
 cudaMemcpy(y_d, y_h, nBytes, cudaMemcpyHostToDevice);

//**********************  inner product ********************************
 cudaThreadSynchronize();
 double tstart = MPI_Wtime();
 icomm.Barrier();                                           // MPI call
 for (int k=0; k<LOOP; ++k) {
    float sum_loc = dscapr_GPU(N, x_d, y_d);                // inner product per MPI process on GPU
    icomm.Allreduce(&sum_loc,&sum,1,MPI::FLOAT,MPI::SUM);   // MPI call: global reduction
}
 cudaThreadSynchronize();
 double tgpu = MPI_Wtime() - tstart;
 cout << "(" << myrank << "/" << numprocs << "): " <<
   "GPU: " << tgpu << " sec.; " <<
   2e-9*N/tgpu << " GFLOPS;  " <<
   2.0*N*sizeof(float)/tgpu/1024/1024/1024 << " GiB/sec.\n ";

 icomm.Barrier();                                           // MPI call
 if (myrank==0)
 {
 if (static_cast<int>(sum)!=N*numprocs)
    {
    cout << " !!  W R O N G  result  " << endl;
    cout << sum << "  vs.  " << N << "(" << ss<< ")" << endl;
    }
 else
    cout << " correct  result  " << endl;
 }

 //******************* free memory *************************************

 delete [] x_h; delete [] y_h;
 cudaFree(x_d); cudaFree(y_d); cudaFree(s_d);

//****************  finish MPI  ****************************************
  MPI::Finalize();
//**********************************************************************

 return 0;
}



