#include <mpi.h>
#include <cassert>
#include <algorithm>    // std::max
#include <iostream>

#include "vecaccu.h"
#include "vdop.h"
#include "communicator.h"

Communicator::Communicator(const MPI::Intracomm &icomm)
    : _icomm(icomm)
{
    //ctor
}

Communicator::~Communicator()
{
    //dtor
}

double Communicator::skalar(const int nd, double x[], double y[]) const
{
    const double s = dscapr(nd, x, y);              // call sequential inner product
    double sg;
    Get_comm().Allreduce(&s, &sg, 1, MPI::DOUBLE, MPI::SUM);

    return (sg);
}



// ---------------------------------------------------------------------
Grid2D_Communicator::Grid2D_Communicator(const MPI::Intracomm &icomm,
        const int procx, const int procy, const int nx, const int ny)
    : 	Communicator(icomm), _procx(procx), _procy(procy), _nx(nx), _ny(ny)
{
    IniGeom(Get_rank(), _procx, _procy, _neigh, _color);

    _nnn[0] = _nx + 1;
    _nnn[1] = _ny + 1;
    _nnn[2] = _nx + 1;
    _nnn[3] = _ny + 1;
    _lnn = std::max(_nx + 1, _ny + 1);
    for (int dir = 0; dir < 4; ++dir) {
        _v_in[dir]  = new double [_lnn];
        _v_out[dir] = new double [_lnn];
    }
}

Grid2D_Communicator::~Grid2D_Communicator()
{
    for (int dir = 0; dir < 4; ++dir) {
        delete [] _v_in[dir];
        delete [] _v_out[dir];
    }
}

void Grid2D_Communicator::VecAccu(const int nd, double w[]) const
{
    assert((_nx + 1) * (_ny + 1) == nd);
    //::VecAccu(_nx, _ny, w, _neigh, _color, Get_comm());     //  finds the VecAccu outside of the class in vecaccu.cpp

// I have to split the communication into the south-north part followed
// by the west-eats part.
// Otherwise the crosspoints are not handled correctly,
// crosspoint on P_0 doesn't get the data from P_3 in case of a 2x2 decomposition.
//
    for (int k = 0; k < 2; ++k) { // k==0: South <--> North,  k==1: East <--> West
        MPI::Request rq[2];

        for (int dir = k; dir < 4; dir += 2) {
            if (_neigh[dir] >= 0) {
                GetBound(dir + 1, _nx, _ny, w, _v_out[dir]);
                rq[dir / 2] = Get_comm().Irecv(_v_in[dir], _nnn[dir], MPI::DOUBLE, _neigh[dir], TAG2);
                Get_comm().Isend(_v_out[dir] , _nnn[dir], MPI::DOUBLE, _neigh[dir], TAG2);
            }
        }

        MPI::Request::Waitall(2, rq);

        for (int dir = k; dir < 4; dir += 2) {
            if (_neigh[dir] >= 0) {
                AddBound(dir + 1, _nx, _ny, w, _v_in[dir]);
            }
        }
    }
}
// ---------------------------------------------------------------------

void GetBounds(const int nx, const int ny, const int nbound, int idx[])
{
    const int //dx = 1,
    dy = nx + 1,
    bl  = 0,
    br  = nx,
    tl  = ny * (nx + 1),
    tr  = (ny + 1) * (nx + 1) - 1;

    int ij = 0;
    // SW
    idx[ij++] = bl;
    // South
    for (int i = bl; i <= br; ++i)
        idx[ij++] = i;
    // SE
    idx[ij++] = br;
    // East
    for (int i = br; i <= tr; i += dy)
        idx[ij++] = i;
    // NE
    idx[ij++] = tr;
    // North
    for (int i = tl; i <= tr; ++i)
        idx[ij++] = i;
    // NW
    idx[ij++] = tl;
    // West
    for (int i = bl; i <= tl; i += dy)
        idx[ij++] = i;

    //std::cout << "### " << nx << "   " << ny << std::endl;
    //std::cout << "::: " << ij << "   " << nbound << std::endl;
    assert(ij == nbound);

    return;
}


Grid2D_gen_Communicator::Grid2D_gen_Communicator(const MPI::Intracomm &icomm,
        const int procx, const int procy, const int nx, const int ny)
    : 	Communicator(icomm), _procx(procx), _procy(procy), _nx(nx), _ny(ny)
{
    const int myid = Get_rank();
    const int ky = myid / procx;
    const int kx = myid % procx;	//    MOD(myid,procx)

    _neigh[1] = (ky == 0)         ?  -1 : myid - procx;	//   South
    _neigh[3] = (kx == procx - 1) ?  -1 : myid + 1;		//   East
    _neigh[5] = (ky == procy - 1) ?  -1 : myid + procx;	//   North
    _neigh[7] = (kx == 0)         ?  -1 : myid - 1;		//   West

    _neigh[0] = (_neigh[7] < 0 || _neigh[1] < 0) ? -1 : _neigh[1] - 1; // SE
    _neigh[2] = (_neigh[1] < 0 || _neigh[3] < 0) ? -1 : _neigh[1] + 1; // SE
    _neigh[4] = (_neigh[3] < 0 || _neigh[5] < 0) ? -1 : _neigh[5] + 1; // SE
    _neigh[6] = (_neigh[5] < 0 || _neigh[7] < 0) ? -1 : _neigh[5] - 1; // SE

    _cnt[0] = 1;           // Vertex
    _cnt[1] = _nx + 1;
    _cnt[2] = 1;           // Vertex
    _cnt[3] = _ny + 1;
    _cnt[4] = 1;           // Vertex
    _cnt[5] = _nx + 1;
    _cnt[6] = 1;           // Vertex
    _cnt[7] = _ny + 1;

    _displ[0] = 0;
    for (int k = 1; k < NPRIM; ++k)
        _displ[k] = _displ[k - 1] + _cnt[k - 1];
    _nbound = _displ[NPRIM-1] + _cnt[NPRIM-1];
    //std::cout << _cnt[0] << " "<< _cnt[1] << " " << _cnt[2] << " " << _cnt[3] << " " << _cnt[4]
    //<< " "<< _cnt[5] << " "<< _cnt[6] << " "<< _cnt[7]   << std::endl;

    _v_in  = new double [_nbound];
    _v_out = new double [_nbound];

    _bound2loc = new int [_nbound];

    GetBounds(_nx, _ny, _nbound, _bound2loc );

    if (Get_rank() == 0) {
		for (int i = 0; i < NPRIM; ++i)  std::cout << "  " << _neigh[i];
        std::cout << std::endl;
        for (int i = 0; i < _nbound; ++i)  std::cout << "  " << _bound2loc[i];
        std::cout << std::endl;
    }

}


Grid2D_gen_Communicator::~Grid2D_gen_Communicator()
{
    delete [] _v_in;
    delete [] _v_out;
    delete [] _bound2loc;
}

void Grid2D_gen_Communicator::VecAccu(const int nd, double w[]) const
{
    //assert((_nx + 1) * (_ny + 1) == nd);

    //for (int i = 0; i < _nbound; ++i) {
        //_v_out[i] = w[_bound2loc[i]];            // GetBound
    //}

    ////MPI::Request rq[NPRIM];                    // now, as member

    //for (int dir = 0; dir < NPRIM; ++dir) { // SW/South/SE/East/NE/North/NW/West

        //if (_neigh[dir] >= 0) {
            //_rq[dir] = Get_comm().Irecv(_v_in + _displ[dir], _cnt[dir], MPI::DOUBLE, _neigh[dir], TAG2);
            //Get_comm().Isend(_v_out + _displ[dir] , _cnt[dir], MPI::DOUBLE, _neigh[dir], TAG2);
        //}
    //}

    //MPI::Request::Waitall(NPRIM, _rq);

    //for (int i = 0; i < _nbound; ++i) {
        //w[_bound2loc[i]] += _v_in[i];           // AddBound
    //}
    Init_VecAccu(nd, w);
    Finish_VecAccu(nd, w);
    return;
}

void Grid2D_gen_Communicator::Init_VecAccu(const int nd, const double r[]) const
{
    assert((_nx + 1) * (_ny + 1) == nd);

    for (int i = 0; i < _nbound; ++i) {
		std::cout << _bound2loc[i] << "  ";
        _v_out[i] = r[_bound2loc[i]];            // GetBound
    }
    std::cout << std::endl;

    //MPI::Request rq[NPRIM];                    // now, as member

    for (int dir = 0; dir < NPRIM; ++dir) { // SW/South/SE/East/NE/North/NW/West

        if (_neigh[dir] >= 0) {
            _rq[dir] = Get_comm().Irecv(_v_in + _displ[dir], _cnt[dir], MPI::DOUBLE, _neigh[dir], TAG2);
            Get_comm().Isend(_v_out + _displ[dir] , _cnt[dir], MPI::DOUBLE, _neigh[dir], TAG2);
        }
    }

    return;
}

void Grid2D_gen_Communicator::Finish_VecAccu(const int, double w[]) const
{
    MPI::Request::Waitall(NPRIM, _rq);

    for (int i = 0; i < _nbound; ++i) {
        w[_bound2loc[i]] += _v_in[i];           // AddBound
    }
    return;	
}

