#ifndef COMMUNICATOR_H
#define COMMUNICATOR_H
#include <mpi.h>
#include "geom.h"

const int TAG2 = 2; //!< Tag to mark the vector accumulation communication

/**
 * Receives data from process with rank @p from into a double-vector @p xout .
 * @param[in] from 	rank of process in communicator @p icomm to receive the message from
 * @param[out] nout 	number of received elements stored in vector @p xout
 * @param[out] xout	vector
 * @param[in] maxbuf	maximal number of elements of  @p xout
 * @param[in] icomm	communicator to use
*/

/**
 * My basis communicator class for providing vector accumulation on interfaces.
 * The derived classes will realized various different communication concepts.
*/
class Communicator {
public:
    /**
     * Initialzes the communicator.
    */
    Communicator(const MPI::Intracomm &icomm);

    Communicator(const Communicator &org) = delete;

    Communicator &operator=(const Communicator &org) = delete;

    /** Default destructor */
    virtual ~Communicator();

    /** Getter;
     * \return Current MPI group;
     */
    const MPI::Intracomm &Get_comm() const
    {
        return _icomm;
    }

    /** Getter;
     * \return My rank in the current MPI group;
     */
    int Get_rank() const
    {
        return _icomm.Get_rank();
    }
    
    /** Getter;
     * \return Size of the current MPI group;
     */
    int Get_size() const
    {
        return _icomm.Get_size();
    }

	/**
	 * Accumulates vector @w.
	 * @param[in]     nd number of elements in vector @w.
	 * @param[in,out] w 	distributed vector will be converted into an accumulated vector
	*/
    virtual void VecAccu(const int nd, double w[]) const = 0;
    
    /**
	 * Inner product of vectors @x and  @y.
	 * @param[in]    nd number of elements in both vectors.
	 * @param[in]    x 	vector
	 * @param[in]    y 	vector
	 * @warning The two vectors must have different types (accumulated/distributed)
	*/
    virtual double skalar(const int nd, double x[], double y[]) const;

protected:
private:
    const MPI::Intracomm _icomm; //!< MPI communication group
};

// ---------------------------------------------------------------------
// 2D Tensor product grid for processes as well as for local grids
class Grid2D_Communicator : public Communicator {
public:
    Grid2D_Communicator(const MPI::Intracomm &icomm,
                        const int procx, const int procy, const int nx, const int ny);

    virtual ~Grid2D_Communicator();

    void VecAccu(const int nd, double w[]) const;

private:
    const int _procx; //!< Number of processes in x-direction
    const int _procy; //!< Number of processes in x-direction
    const int _nx;    //!< Number of local intervals in x-direction
    const int _ny;    //!< Number of local intervals in y-direction
    int _neigh[4];    //!< MPI-ID of neighboring processes in south/east/north/west direction
    int _color;       //!< my black/white color

    //  temporary arrays for communication
    int _nnn[4];      //!< Number of data in this direction
    int _lnn;         //!< max. number of data in one direction
    mutable double *_v_in[4];  //!< array of pointers for receiving data
    mutable double *_v_out[4]; //!< array of pointers for sending data
};

// ---------------------------------------------------------------------
// Communicator originates from a 2D Tensor product grid but then we'll forget
// about this in the vector accumulation.

// https://www.open-mpi.org/doc/current/man3/MPI_Alltoallv.3.php

class Grid2D_gen_Communicator : public Communicator {
public:
    Grid2D_gen_Communicator(const MPI::Intracomm &icomm,
                            const int procx, const int procy, const int nx, const int ny);

    virtual ~Grid2D_gen_Communicator();

    //int Get_nBound() const {return _nbound;}

    void VecAccu(const int nd, double w[]) const;
    
    //
    /**
	 * Starts a non-blocking accumulation with vector @r.
	 * Vector @r will be copied to the output buffer and the non-blocking MPI-send/recv are started.
	 * @param[in]     nd number of elements in vector @r.
	 * @param[in]     r  distributed vector with data for accumulation
	*/
    void Init_VecAccu(const int nd, const double r[]) const;
    
    /**
	 * Finishes a non-blocking accumulation with vector @w.
	 * Checks whether the non-blocking MPI-recv have been finished and adds 
	 * the communicated values to the appropriate entries in w.
	 * @param[in]     nd number of elements in vector @w.
	 * @param[in,out] w  accumulated vector with data for accumulation
	*/
    void Finish_VecAccu(const int nd, double w[]) const;
    

private:
    static const int NPRIM = 8;
    const int _procx; //!< Number of processes in x-direction
    const int _procy; //!< Number of processes in x-direction
    const int _nx;    //!< Number of local intervals in x-direction
    const int _ny;    //!< Number of local intervals in y-direction
    int _neigh[NPRIM];//!< MPI-ID of the neighboring process in south/SE/east/NE/north/NW/west/SW direction
    int _cnt[NPRIM];  //!< Number of data in this direction
    int _displ[NPRIM];//!< Index  where the data for this direction start in a sequential vector.
    int _nbound;      //!< Number of boundary data
    int *_bound2loc;  //!< maps sequential bound index onto local node numbering

    ////  temporary arrays for communication
    mutable double *_v_in;   //!< array for receiving data
    mutable double *_v_out;  //!< array for sending data
    mutable MPI::Request _rq[NPRIM];  //!< stores the status of all requests for  o n e  communication
};


#endif // COMMUNICATOR_H
