#include <mpi.h>
#include <cassert>
#include <algorithm>    // std::max

#include "vecaccu.h"
#include "vdop.h"
#include "communicator.h"

Communicator::Communicator(const MPI::Intracomm &icomm)
    : _icomm(icomm)
{
    //ctor
}

Communicator::~Communicator()
{
    //dtor
}

double Communicator::skalar(const int nd, double x[], double y[]) const
{
    const double s = dscapr(nd, x, y);              // call sequential inner product
    double sg;
    Get_comm().Allreduce(&s, &sg, 1, MPI::DOUBLE, MPI::SUM);

    return (sg);
}



// ---------------------------------------------------------------------
Grid2D_Communicator::Grid2D_Communicator(const MPI::Intracomm &icomm,
        const int procx, const int procy, const int nx, const int ny)
    : 	Communicator(icomm), _procx(procx), _procy(procy), _nx(nx), _ny(ny)
{
    IniGeom(Get_rank(), _procx, _procy, _neigh, _color);
       
    _lnn = std::max(_nx + 1, _ny + 1);
    for (int dir = 0; dir < 4; ++dir) {
        _v_in[dir]  = new double [_lnn];
        _v_out[dir] = new double [_lnn];
    }
}

Grid2D_Communicator::~Grid2D_Communicator()
{
    for (int dir = 0; dir < 4; ++dir) {
        delete [] _v_in[dir];
        delete [] _v_out[dir];
    }
}

void Grid2D_Communicator::VecAccu(const int nd, double w[]) const
{
    assert((_nx + 1) * (_ny + 1) == nd);
    //::VecAccu(_nx, _ny, w, _neigh, _color, Get_comm());     //  finds the VecAccu outside of the class in vecaccu.cpp
    
// I have to split the communication into the south-north part followed
// by the west-eats part.
// Otherwise the crosspoints are not handled correctly,
// crosspoint on P_0 doesn't get the data from P_3 in case of a 2x2 decomposition.
//
    for (int k = 0; k < 2; ++k) { // k==0: South <--> North,  k==1: East <--> West
        MPI::Request rq[2];

        for (int dir = k; dir < 4; dir += 2) {
            if (_neigh[dir] >= 0) {
                GetBound(dir + 1, _nx, _ny, w, _v_in[dir]);
                rq[dir/2] = Get_comm().Irecv(_v_out[dir], _nnn[dir], MPI::DOUBLE, _neigh[dir], TAG2);
                Get_comm().Isend(_v_in[dir] , _nnn[dir], MPI::DOUBLE, _neigh[dir], TAG2);
            }
        }

        MPI::Request::Waitall(2, rq);

        for (int dir = k; dir < 4; dir += 2) {
            if (_neigh[dir] >= 0) {
                AddBound(dir + 1, _nx, _ny, w, _v_out[dir]);
            }
        }
    }    
}

// ---------------------------------------------------------------------
//Grid2D_gen_Communicator::Grid2D_gen_Communicator(const MPI::Intracomm &icomm,
        //const int procx, const int procy, const int nx, const int ny)
    //: 	Communicator(icomm), _procx(procx), _procy(procy), _nx(nx), _ny(ny)
//{
	////int tmp;
    ////IniGeom(Get_rank(), _procx, _procy, _neigh, tmp);
    
    ////for (int dir = 3; dir >=0; --dir)
    ////{
		////_neigh[2*dir] = _neigh[dir];
	////}
	
	//const int myid = Get_rank();
	//const int ky = myid / procx;
    //const int kx = myid % procx;	//    MOD(myid,procx)

    //_neigh[0] = (ky == 0)         ?  -1 : myid - procx;	//   South
    //_neigh[2] = (kx == procx - 1) ?  -1 : myid + 1;		//   East
    //_neigh[4] = (ky == procy - 1) ?  -1 : myid + procx;	//   North
    //_neigh[6] = (kx == 0)         ?  -1 : myid - 1;		//   West
    
    //_neigh[1] = (_neigh[0]<0 || _neigh[2]<0) ? -1 : _neigh[0]-1; // SE
    //_neigh[3] = (_neigh[2]<0 || _neigh[4]<0) ? -1 : _neigh[4]-1; // SE
    //_neigh[5] = (_neigh[4]<0 || _neigh[6]<0) ? -1 : _neigh[4]+1; // SE
    //_neigh[7] = (_neigh[6]<0 || _neigh[0]<0) ? -1 : _neigh[0]+1; // SE

    //_cnt[0] = _nx + 1;
    //_cnt[1] = 1;           // Vertex
    //_cnt[2] = _ny + 1;
    //_cnt[3] = 1;           // Vertex
    //_cnt[4] = _nx + 1;
    //_cnt[5] = 1;           // Vertex
    //_cnt[6] = _ny + 1;
    //_cnt[7] = 1;           // Vertex
    
    //_displ[0] = 0
    //for (int k = 1; k<8; ++k)  _displ[k] = _displ[k-1]+_cnt[0];
    //_nbound = _displ[7]+_cnt[7];
    
    
    
    //_lnn = std::max(_nx + 1, _ny + 1);
    //for (int dir = 0; dir < 4; ++dir) {
        //_v_in[dir]  = new double [_lnn];
        //_v_out[dir] = new double [_lnn];
    //}
//}


