//		MPI code in C++.
//		See [Gropp/Lusk/Skjellum, "Using MPI", p.33/41 etc.]
//		and  /opt/mpich/include/mpi2c++/comm.h  for details

#include <iostream>
#include <cmath>
using namespace std;
#include <mpi.h>
#include "debugd.h"
#include "vecaccu.h"
#include "geom.h"
#include "userset.h"
#include "jacsolve.h"
#include "getmatrix.h"

int main(int argc, char* argv[])
{
    const bool bdebug=false;		// debugging at the end??

    MPI::Init(argc,argv);
    MPI::Intracomm icomm(MPI::COMM_WORLD);
    int numprocs = icomm.Get_size();
    int myrank   = icomm.Get_rank();

    if (myrank == 0)
    { cout << "\n There are " << numprocs << " processes running.\n \n"; }
    icomm.Barrier();

    const int procx = static_cast<int>(sqrt(numprocs+0.0));
    const int procy = procx;

    if (procy * procx != numprocs)
    {
      cout << "\n Wrong number of processors !\n \n";
    }
    else
    {
      int neigh[4], color;
      IniGeom(myrank, procx, procy, neigh, color); // Topology of domain decomposition
      icomm.Barrier();
      double xl,xr,yb,yt;                          // subdomain coordinates
      IniCoord(myrank, procx, procy, xl, xr, yb, yt);

      bool ScaleUp = !true;
      int nx,ny,NXglob,NYglob;
      /* number of local intervals on (xl,xr)=:nx, (yb,yt)=:ny */
      if (ScaleUp)
      {
      /*		for scale up evaluation   */
        if (myrank==0) cout << "scale up evaluation\n";
        nx = 10;
        ny = 10;
        NXglob = nx*procx;
        NYglob = ny*procy;
      }
      else
      {
      /*		for speed up evaluation   */
        if (myrank==0) cout << "speed up evaluation\n";
        NXglob = NYglob = 40;
        nx = NXglob/procx;
        ny = NYglob/procy;
      }

// Initialze the mesh
      int nnode, nelem;        // number of ccordinates, number of triangular elements
      double *xc;              // coordinates         xc[nnode,2]
      int    *ia;              // element conectivity ia[nelem,3]
      GetMesh(nx, ny, xl, xr, yb, yt, nnode, xc, nelem, ia);
      //DebugMesh(nnode, xc, nelem, 3, ia);  {int ijk; cin >> ijk;}

// Matrix Graph
      int nnz(-1);                           // number of non-zero eelemnts in matrix
      double *sk;                            // sk[nnz]  non-zero entries
      int    *ik,                            // ik[nnz]  column indices
             *id;                            // id[nnode+1] index of row begin
      Get_Matrix_Pattern(nelem, 3, ia, nnz, id, ik, sk);
      //cout << "AA " << nx << "  " << ny << "  " << nnode << "  "<< nnz << endl;
      //DebugMatrix(nnode, id, ik, sk);  {int ijk; cin >> ijk;}

      /* allocate memory */
      double *u = new double [nnode],          /* temperature       */
             *f = new double [nnode];          /* r.h.s.            */

    /* matrix, rhs generation including Dirichlet B.C. */
      GetMatrix (nelem, 3, ia, nnode, xc, nnz, id, ik, sk, f);
      //DebugMatrix(nnode, id, ik, sk);  {int ijk; cin >> ijk;}

      SetU(nx, ny, u);			     /* set solution/boundary values */
      ApplyDirichletBC(nx, ny, neigh, u, id, ik, sk, f); /* apply B.C.       */
//       DebugMatrix(nnode, sk, id, ik);

//      double tstart = omp_get_wtime();                       // OpenMP timing
      double tstart = MPI_Wtime();                             // MPI timing

      JacobiSolve(nnode, nx, ny, neigh, color, icomm, id, ik, sk, f, u );

      //double t1 = omp_get_wtime() - tstart;                 // OpenMP timing
      double t1 = MPI_Wtime() - tstart;                       // MPI timing
      if (myrank == 0) {cout << "WTIME(MPI) timing in sec. : " << t1 << endl;}

      SaveVectorP(myrank, "t.dat", u, nx, ny, xl, xr, yb, yt);

      if (bdebug)
      {
        /*  simple check of CrsMult function	*/
     	/* check matrix, rhs */
        DebugMatrix(nnode, id, ik, sk);
        DebugD(nnode, f);

        for (int i=0; i<nnode; ++i)  u[i] = i;

	    CrsMult(f, u, nnode, id, ik, sk);
        DebugD(nnode,f);

	/*  simple check of VecAccu function	*/
	/* initialize u  */
        for (int i=0; i<nnode; ++i)  u[i] = myrank*100+i;

        DebugD(nnode,u);
        VecAccu(nx,ny,u,neigh,color,icomm);
        DebugD(nnode,u);
      }

      delete [] id; delete [] ik; delete [] sk; delete [] f; delete [] u;
  }

  MPI::Finalize();
}


