#ifndef GETMATRIX_FILE
#define GETMATRIX_FILE


/**
      * Determines the matrix graph of an CRS matrix (@p id, @p ik, @p sk) based on the
      * element connectivity @p ia.

        The memory for arrays .@p id, @p ik, @p sk is allocated inside the function.

      * @param[in]  nelem   number of triangular elements
      * @param[in]  ndof_e  number of f.e. nodes per element
      * @param[in]  ia      element connectivity array[nelem][ndof_e] for linear triangular elements
      * @param[out] nnz         number of non-zero elements in matrix
      * @param[out] id          array[nnode+1] storing indices of first entry of a row
      * @param[out] ik[nnz]     array storing column indices of non-zero entries
      * @param[out] sk[nnz]     array storing values of non-zero entries
*/
void Get_Matrix_Pattern(const int nelem, const int ndof_e, int const *const ia,
                        int &nnz, int* &id, int* &ik, float* &sk);



/**
 * Calculates the element stiffness matrix @p ske and the element load vector @p fe
 * of one triangular element with linear shape functions.
 * @param[in]	ial	node indices of the three element vertices
 * @param[in]	xc	vector of node coordinates with x(2*k,2*k+1) as coodinates of node k
 * @param[out] ske	element stiffness matrix
 * @param[out] fe	element load vector
 */
void CalcElem(const int ial[3], const float xc[], float ske[3][3], float fe[3]);

/**
 * Adds the element stiffness matrix @p ske and the element load vector @p fe
 * of one triangular element with linear shape functions to the appropriate positions in
 * the symmetric stiffness matrix, stored as CSR matrix K(@p sk,@p id, @p ik)
 * @param[in] ial   node indices of the three element vertices
 * @param[in] ske	element stiffness matrix
 * @param[in] fe	element load vector
 * @param[out] sk	vector non-zero entries of CSR matrix
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @param[out] f	distributed local vector storing the right hand side
*/
void AddElem(const int ial[3], const float ske[3][3], const float fe[3],
	const int id[], const int ik[], float sk[], float f[]);


/**
 * Print entries of the CSR matrix K(@p sk,@p id, @p ik).
 * @param[in] nnode	number of rows
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @param[in] sk	vector non-zero entries of CSR matrix
*/
void DebugMatrix(const int nnode, const int id[], const int ik[], const float sk[]);

void DebugVector(const int nnode, const float v[]);


/**
 * Computes the stiffness matrix  in CRS format(@p id, @p ik, @p sk) and load vector @p f for the Poisson problem
 * without incorporating the boundary conditions
 * The matrix pattern has to be generated previously and the memery has to be allocated.
 *
 * @see Get_Matrix_Pattern
 *
 * @param[in] nelem	number of finite elements
 * @param[in] ndof_e	degrees of freedom per element
 * @param[in] ia	element connectivity array[nelem][ndof_e]
 * @param[in] nnode	number of coordinates (== number of unknowns)
 * @param[in] xc	coordinates[nnode][2]
 * @param[in] nnz	number of non-zero entries in CRS matrix
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @param[out] sk	vector of non-zero entries of CSR matrix
 * @param[out] f	distributed local vector storing the right hand side
 */
void GetMatrix (const int nelem, const int ndof_e, const int ia[], const int nnode, const float xc[],
                const int nnz, const int id[], const int ik[], float sk[], float f[]);



/**
 * Applies Dirichlet boundary conditions to symmetric CSR matrix  (@p sk,@p id, @p ik) and right hand side @p f
 * of a system of equations resulting from a FEM/FDM discretization
 * @param[in] nx	number of discretization intervals in x-direction
 * @param[in] ny	number of discretization intervals in y-direction
 * @param[in] neigh	vector of 4 elements containing the ranks of neighbouring processes, -1 indicates a boundary in this direction
 * @param[in] u		local vector with appropriately initialized b.c.
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @param[inout] sk	vector non-zero entries of CSR matrix
 * @param[inout] f	right hand side
*/
void ApplyDirichletBC(const int nx, const int ny, const int neigh[],
       const float u[], const int id[], const int ik[] , float sk[], float f[]);


/**
 * Finds in a CRS matrix the access index for an entry at  row @p row  and column @p col.
 *
 * @param[in] row	row index
 * @param[in] col	column index
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @return index for element (@p row, @p col). If no appropriate entry exists then -1 will be returned.
 *
 * @warning assert() stops the function in case that  element (@p row, @p col) doesn't exist.
*/
int fetch(const int row, const int col, const int id[], const int ik[]);


/**
 * Calculates the defect  w := f - K*u  a CSR matrix K(@p id, @p ik, @p sk).
 *
 * @param[in,out] w	distributed local vector.
 * @param[in] f		distributed local vector.
 * @param[in] u		accumulated local vector.
 * @param[in] nnode	number of rows
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @param[in] sk	vector non-zero entries of CSR matrix
*/
//#pragma acc routine gang
void Defect(float w[], const float f[], const float u[],
            const int nnode, const int id[], const int ik[], const float sk[]);


/**
 * Multiplies vector @p u with a CSR matrix K(@p id, @p ik, @p sk).
 * w := K*u
 * @param[in,out] w	distributed local vector.
 * @param[in] u		accumulated local vector.
 * @param[in] nnode	number of rows
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @param[in] sk	vector non-zero entries of CSR matrix
*/
void CrsMult(float w[], const float u[], const int nnode, const int id[], const int ik[], const float sk[]);



/**
 * Copies the diagonal entries of a CSR matrix (@p sk,@p id, @p ik)
 * onto a vector @p d.
 * @param[in] nnode	number of rows
 * @param[in] id	index vector containing the first entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @param[in] sk	vector non-zero entries of CSR matrix
 * @param[out] d	accumulated diagonal entries of CSR matrix
*/
void GetDiag(const int nnode, const int id[], const int ik[], const float sk[], float d[]);



#endif

