#ifndef VDOP_FILE
#define VDOP_FILE

/** @brief  Copies elements from vector @p yp onto vector @p xp.
 *  A stride @p iy =0 for the source stride copies a constant value stored in @p y[0].
 *
 * @param[in]  n  number of elements to copy
 * @param[out] xp target vector
 * @param[in]  ix stride for target vector @p xp
 * @param[in]  yp source vector
 * @param[in]  iy stride for source vector @p yp
 *
 */
void vdcopy(const int n, double xp[], const int ix,
                   const double yp[], const int iy);

/** @brief  Vector addition x = y+z.
 *  A stride @p iy =0 for the source stride adds a constant value stored in @p y[0].
 *  The same holds for vector @p z.
 *
 * @param[in]  n  number of elements to add
 * @param[out] x  target vector
 * @param[in]  ix stride for target vector @p x
 * @param[in]  y  source vector
 * @param[in]  iy stride for source vector @p y
 * @param[in]  z  source vector
 * @param[in]  iz stride for source vector @p z
 *
 */
void vdplus(const int n, double x[], const int ix,
                   const double y[], const int iy,
                   const double z[], const int iz);

/** @brief  Element-wise vector divison x = y/z.
 *  A stride @p iy =0 for the source stride adds a constant value stored in @p y[0].
 *  The same holds for vector @p z.
 *
 * @param[in]  n  number of elements to divide
 * @param[out] x  target vector
 * @param[in]  ix stride for target vector @p x
 * @param[in]  y  source vector
 * @param[in]  iy stride for source vector @p y
 * @param[in]  z  source vector
 * @param[in]  iz stride for source vector @p z
 *
 */
void vddiv(const int n,  double x[], const int ix,
                   const double y[], const int iy,
                   const double z[], const int iz);

/** @brief  Element-wise vector multiplication x = y*z.
 *  A stride @p iy =0 for the source stride adds a constant value stored in @p y[0].
 *  The same holds for vector @p z.
 *
 * @param[in]  n  number of elements to multiply
 * @param[out] x  target vector
 * @param[in]  ix stride for target vector @p x
 * @param[in]  y  source vector
 * @param[in]  iy stride for source vector @p y
 * @param[in]  z  source vector
 * @param[in]  iz stride for source vector @p z
 *
 */
void vdmult(const int n, double x[], const int ix,
                   const double y[], const int iy,
                   const double z[], const int iz);

/** @brief  Element-wise daxpy opration x = y + alfa*z.
 *  The stride is one.
 *
 * @param[in]  n  number of elements to operate
 * @param[out] x target vector
 * @param[in]  y source vector
 * @param[in]  a scalar value
 * @param[in]  z source vector
 *
 */
void vdaxpy(const int n, double x[], const double y[],
                   const double a,   const double z[] );


/** @brief  Calculates the Euclidian inner product of two vectors.
 *
 * @param[in]  n  number of elements in both vectors.
 * @param[in]  x source vector
 * @param[in]  y source vector
 * @return Euclidian inner product @f$\langle x,y \rangle@f$
 *
 */
double dscapr(const int n, const double x[], const double y[]);


/** @brief  Copies elements from vector @p yp onto vector @p xp.
 *  A stride @p iy =0 for the source stride copies a constant value stored in @p y[0].
 *
 * @param[in]  n  number of elements to copy
 * @param[out] xp target vector
 * @param[in]  ix stride for target vector @p xp
 * @param[in]  yp source vector
 * @param[in]  iy stride for source vector @p yp
 *
 */
void vicopy(const int n, int xp[], const int ix,
                   const int yp[], const int iy);


#endif
